{   
   MultiProcA.fth
   automatic multiProcessing of Tasks
   inspired by multiprocessor Forth Kernel page 14 in:
   http://www.forth.org/fd/FD-V11N3.pdf
   
   uses AutoS.fth
   11.07.23 Christof Eberspaecher 
}

!polls \ clear multitasking

IFDEF *Tests*	
    oldorgT org
    oldLutPC lutPC !
    oldLutOrg lutOrg !
    FORGET *Tests*	 }
pub *Tests*	    PRINT" P2 Test" ;

0 bytes oldorgT
lutPC @ := oldLutPC 
lutOrg @ := oldLutOrg

--- LOCKS v1 for Taqoz Reloaded v2.8 - Bob Edwards May 2021

--- If two or more cogs are writing to the same data in hub memory, that data would be in jeopardy from race conditions as each cog performs
---  its read-modify-write cycle. The outcome of two cogs writing to the same address at very nearly the same time is unknown - which of the
--- two values ends up at the address?

--- To fix that, the P2 has a pool of 16 semaphore bits called locks....

--- Allocate a lock, returning the lock number n. If n = 0-15, lock was allocated, if n=-1 then all locks were already allocated 
code LOCKNEW ( -- n )
			>PUSHX				'make space on the stack
			locknew a wc		't.o.s. new lock allocation number
	if_nc	ret					'successful allocation
	_ret_	mov a,#-1			'else signal all locks taken
end

--- Return lock number n (0-15) to the pool
code LOCKRET ( n -- )
			lockret a
			DROP;
			ret
end

--- test LOCKNEW & LOCKRET
\ CRLF 17 FOR LOCKNEW . CRLF NEXT								--- try allocating one too many locks
\ 13 LOCKRET CRLF ." Did we release lock 13? - " LOCKNEW .	--- check we can release a lock
\ CRLF 16 FOR I LOCKRET NEXT									--- return all locks

--- Attempt to 'take' Lock n, flag = 0 if successful, else flag = 1
code LOCKTRY	( n -- flag )
			locktry a wc
	_ret_	wrnc a				' a = carry flag
end

--- Release Lock n (0-15) - only the cog that took the lock is permitted to do this
code LOCKREL	( n -- )
			lockrel a
			DROP;
end

--- Read lock n status, lock_status 1 = unlocked, 0 = locked - N.B. if lock is not owned, results invalid
code LOCK?		( n -- lock_owner lock_status  )
			>PUSHX				' make room for status
			lockrel b wc		' t.o.s.+1   = cog no. whose lock it is
	_ret_	wrnc a				' t.o.s.     = lock status
end

long taskLock
locknew taskLock !

0 value nextReady#
16 words readyTasks
readyTasks 16 2* erase

: +ready {: cfa# , tnr# -- } \ add a new task to ready list
   begin taskLock @ locktry 0= until
   -1 to_tnr#
   begin
      1 +to_tnr#
   readyTasks tnr# 2* + w@ 0= until \ empty?
   cfa# readyTasks tnr# 2* + w!  
   taskLock @ lockrel
; forgetLocals

: getTask (  -- cfa# ) \ from readyTasks, zero for found none
   begin taskLock @ locktry 0= until
   1 +to nextReady#
   nextReady# 16 = if 0 to nextReady# then
   readyTasks nextReady# 2* + w@ \ get cfa
   0 readyTasks nextReady# 2* + w! \ clear list position 
   taskLock @ lockrel   
;

: worker \ work one task
   getTask ?dup if
      dup execute
      +ready \ reinstall into list
   then
;

: workerCog \ usage: ' workerCog 6 run 
   getLutCode
   toInit
   begin
      worker
   again
;

\ EXAMPLE ###################################################

8 longs counters
counters 8 cells erase

: cnt0
   1000 us \ waste some time
   counters 0 cells + ++ ;

: cnt1
   1000 us
   counters 1 cells + ++ ;

: cnt2
   1000 us
   counters 2 cells + ++ ;

: cnt3
   1000 us
   counters 3 cells + ++ ;

: cnt4
   1000 us
   counters 4 cells + ++ ;

' cnt0 +ready
' cnt1 +ready
' cnt2 +ready
' cnt3 +ready
' cnt4 +ready

: .cnt
   crlf
   8 for
      counters i cells + @ .
   next
;

: meas
   counters @ \ read 1st counter
   lap
   begin
   dup counters @ swap - 1000 > until
   lap .lap drop
;

' workerCog 6 run meas
' workerCog 5 run meas
' workerCog 4 run meas
' workerCog 3 run meas
' workerCog 2 run meas
' workerCog 1 run meas
meas



{
TAQOZ# ' workerCog 6 run meas --- 1,055,593,884 cycles= 4,294,921,519ns @200MHz ok
TAQOZ# ' workerCog 5 run meas --- 532,473,868 cycles= 2,662,369,340ns @200MHz ok
TAQOZ# ' workerCog 4 run meas --- 361,353,452 cycles= 1,806,767,260ns @200MHz ok
TAQOZ# ' workerCog 3 run meas --- 278,938,236 cycles= 1,394,691,180ns @200MHz ok
TAQOZ# ' workerCog 2 run meas --- 237,404,668 cycles= 1,187,023,340ns @200MHz ok
TAQOZ# ' workerCog 1 run meas --- 222,407,260 cycles= 1,112,036,300ns @200MHz ok
TAQOZ# meas --- 222,472,684 cycles= 1,112,363,420ns @200MHz ok
}

   